﻿Imports System.Threading
Imports GrapeCity.CalcEngine

Namespace SpreadNETZetaCalculator
    ''' <summary>
    ''' Class implementing the custom fuction to calculate the Riemann ZETA function for a set of points along a vertical line (e.g. the critical line x = 1/2).
    ''' </summary>
    ''' <remarks>
    ''' This implementation calculates the analytic continuation ζ(s) in the complex plane, which is defined for s > 0.
    ''' Usage of this function is as follows:
    ''' ZETA(re, im, count, incr, iter) where:
    ''' re = the real component of the initial point to calculate (must be > 0)
    ''' im = the imaginary component of the initial point to calculate (values are symetric about x-axis => ζ(x+bi) = ζ(x-bi) )
    ''' count = the number of points to calculate
    ''' incr = the increment between each point to calculate
    ''' iter = the number of iterations to calculate -- the higher this number, the more accurate the calculation (and longer it will take)
    ''' </remarks>
    Friend NotInheritable Class ZetaFunction
        Inherits [Function]
        ' keep count of function calls
        Friend Shared cnt As Integer = 0
        ' time calculation
        Friend Shared t1 As DateTime
        Friend Shared t2 As DateTime
        ''' <summary>
        ''' Create the ZetaFunction.
        ''' </summary>
        Public Sub New()
            MyBase.New("ZETA", 4, 5, FunctionAttributes.Array)
        End Sub
        ''' <summary>
        ''' Evaluate the ZETA function.
        ''' </summary>
        ''' <param name="args">IArguments specifying arguments passed to ZETA</param>
        ''' <param name="result">IValue for returning calculation result</param>
        Protected Overrides Sub Evaluate(ByVal args As IArguments, ByVal result As IValue)
            cnt += 1 ' increment function call counter
            t1 = DateTime.Now ' start timing evaluation
            Dim l As Integer = 10000 ' default to 10K iterations if not specified
            Dim d As Double = args(0).GetPrimitiveValue(args.EvaluationContext).GetNumber() ' real part
            Dim d1 As Double = args(1).GetPrimitiveValue(args.EvaluationContext).GetNumber() ' imaginary part
            Dim count As Integer = CInt(args(2).GetPrimitiveValue(args.EvaluationContext).GetNumber()) ' number of points to calculate
            Dim incr As Double = args(3).GetPrimitiveValue(args.EvaluationContext).GetNumber() ' increment between points
            ' iterations is optional 
            If args.Count > 4 Then l = CInt(args(4).GetPrimitiveValue(args.EvaluationContext).GetNumber())
            Dim array As IValuesArray = result.CreateValuesArray(CInt(count), 2) ' create array to return values
            Dim arr As Double(,) = New Double(count - 1, 1) {} ' create local array to calculate values (can't set values to IValuesArray from another thread)
            '  do the ZETA calculation in a Task using Parallel.For for better performance
            Dim t As Task = New Task(
                Sub()
                    Parallel.[For](0, count, Sub(i)
                        Dim d2 As Double = 0.0
                        Dim d3 As Double = 0.0
                        Dim d4 As Double = 0.0
                        Dim d5 As Double = 0.0
                        Dim dd1 As Double = d1 + i * incr

                        For l1 As Integer = 1 To l ' calculate Dirichlet Eta series partial sum of l iterations
                            Dim d6 As Double = 2 * l1 - 1
                            Dim d7 As Double = 2 * l1
                            Dim d14 As Double = 0.0
                            Dim d15 As Double = 0.0
                            Dim d16 As Double = 0.0
                            Dim d8 As Double = (dd1 * Math.Log(d6)) Mod 6.2831853071795862
                            Dim d9 As Double = (dd1 * Math.Log(d7)) Mod 6.2831853071795862
                            Dim d10 As Double = -d
                            Dim d11 As Double = Math.Pow(d6, d10)
                            Dim d12 As Double = Math.Pow(d7, d10)
                            d2 += d11 * Math.Cos(d8) - d12 * Math.Cos(d9)
                            d3 += d12 * Math.Sin(d9) - d11 * Math.Sin(d8)
                            d4 += Math.Pow(d6, -0.5) * Math.Cos(d8) - Math.Pow(d7, -0.5) * Math.Cos(d9)
                            d14 = Math.Pow(d7, -0.5) * Math.Sin(d9)
                            d15 = Math.Pow(d6, -0.5)
                            d16 = Math.Sin(d8)
                            d5 += d14 - d15 * d16
                        Next
                        ' calculate Zeta from Eta value
                        Dim d17 As Double = 1.0 - Math.Pow(2, 1.0 - d) * Math.Cos(d1 * Math.Log(2))
                        Dim d18 As Double = Math.Pow(2, 1.0 - d) * Math.Sin(d1 * Math.Log(2))
                        Dim d19 As Double = (d17 * d2 + d18 * d3) / (Math.Pow(d17, 2) + Math.Pow(d18, 2))
                        Dim d20 As Double = (d17 * d3 - d18 * d2) / (Math.Pow(d17, 2) + Math.Pow(d18, 2))
                        Interlocked.Exchange(arr(i, 0), d19)
                        Interlocked.Exchange(arr(i, 1), d20)
                    End Sub)
                End Sub)
            t.RunSynchronously() ' need to run this task synchronously on the UI thread
            ' copy values from arr to the IValuesArray
            For i As Integer = 0 To count - 1
                array(i, 0).SetValue(arr(i, 0))
                array(i, 1).SetValue(arr(i, 1))
            Next
            ' set result -- this is important! If you don't return a result, then the cells remain marked dirty and Evaluate will be called again...
            result.SetValue(array)
            t2 = DateTime.Now
        End Sub
    End Class
End Namespace
